/*
 * All files in the distribution of JTB, The Java Tree Builder are
 * Copyright 1997, 1998 by the Purdue Research Foundation of Purdue
 * University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and
 * other materials related to such distribution and use acknowledge
 * that the software was developed at Purdue University, West Lafayette,
 * Indiana by Kevin Tao and Jens Palsberg.  No charge may be made
 * for copies, derivations, or distributions of this material
 * without the express written consent of the copyright holder.
 * Neither the name of the University nor the name of the author
 * may be used to endorse or promote products derived from this
 * material without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 */

package EDU.purdue.jtb.misc;

import EDU.purdue.jtb.visitor.Printer;
import EDU.purdue.jtb.syntaxtree.*;
import java.util.*;
import java.io.*;

/**
 * Class ClassInfo is used by the visitors to store information about a class
 * including its name, a list of field types, and a list of field names.
 *
 * Class CommentPrinter appears at the end of this file.
 */
public class ClassInfo {
   private Node astNode;
   private String name;
   private Vector typeList;
   private Vector nameList;

   public ClassInfo(Node node, String n) {
      astNode = node;
      name = n;
      typeList = new Vector();
      nameList = new Vector();
   }

   public Node   getAstNode()  { return astNode; }
   public String getName()     { return name; }
   public Vector getTypeList() { return typeList; }
   public Vector getNameList() { return nameList; }

   public void addField(String type, String name) {
      typeList.addElement(type);
      nameList.addElement(name);
   }

   //
   // Returns the EBNF production of the current class as a bunch of comments
   // showing which field names belong to which parts of the production.
   //
   public String getEbnfProduction(Spacing space) {
      StringWriter buf = new StringWriter();
      PrintWriter out = new PrintWriter(buf);
      String temp;
      BufferedReader str;

      // print result into buffer
      astNode.accept(new CommentPrinter(out, space));
      out.flush();
      temp = buf.toString();
      return temp;
   }

   //
   // Returns a string representing the class.
   //
   public String getClassString(Spacing space) {
      Enumeration types = typeList.elements();
      Enumeration names = nameList.elements();
      StringWriter buf = new StringWriter();
      PrintWriter out = new PrintWriter(buf);

      //
      // Output class definition
      //
      if ( Globals.superClassname == null )
         out.println(space.spc + "public class " + name + " implements Node {");
      else
         out.println(space.spc + "public class " + name + " extends "+ Globals.superClassname + " implements Node {");

      space.updateSpc(+1);

      //
      // Output data fields
      //
      out.println(space.spc + "private Node parent;");
      for ( ; types.hasMoreElements(); )
         out.println(space.spc + "public " + (String)types.nextElement() + " " +
                     (String)names.nextElement() + ";");

      //
      // Output constructor header
      //
      out.println();
      out.print(space.spc + "public " + name + "(");
      
      types = typeList.elements();
      if ( types.hasMoreElements() )
         out.print((String)types.nextElement() + " n0");

      for ( int count = 1; types.hasMoreElements(); ++count )
         out.print(", " + (String)types.nextElement() + " n" + count);

      out.println(") {");

      //
      // Output constructor body
      //
      names = nameList.elements();
      space.updateSpc(+1);

      for ( int count = 0; names.hasMoreElements(); ++count ) {
         String name = (String)names.nextElement();
         out.println(space.spc + name + " = n" + count + ";");
         out.println(space.spc + "if ("+name+"!=null) "+ name + ".setParent(this);");
      }

      space.updateSpc(-1);
      out.println(space.spc + "}");

      //
      // Output visit method
      //
      out.println();
      out.println(space.spc + "public void accept(" + Globals.visitorPackage +
                  ".Visitor v) {");
      space.updateSpc(+1);
      out.print(space.spc + "v.visit");
      if ( Globals.oldVisitNames ) out.print(name);
      out.println("(this);");
      space.updateSpc(-1);
      out.println(space.spc + "}");

      //
      // Output get/set parent methods
      //
      out.println();
      out.println(space.spc+"public void setParent(Node n) { parent = n; }");
      out.println(space.spc+"public Node getParent() {");
      out.println(space.spc+"   if (parent == null) throw new Error(\"Parent not set.\");");
      out.println(space.spc+"   return parent;");
      out.println(space.spc+"}");

      //
      // Output set children method
      //
      out.println(space.spc+"public void setChildAt(Node n, int i) {");

      names = nameList.elements();
      types = typeList.elements();

      space.updateSpc(+1);

      for (int count =0 ; names.hasMoreElements(); count ++) {
        if (count ==0) {
          out.println(space.spc+"if (i == "+count+")");
          out.println(space.spc+"   "+(String)names.nextElement()+" = ("+(String)types.nextElement()+") n;"); 
        }
        else {
          out.println(space.spc+"else if (i == "+count+")");
          out.println(space.spc+"   "+(String)names.nextElement()+" = ("+(String)types.nextElement()+") n;"); 
        }
      }
      
      out.println(space.spc+"else");
      out.println(space.spc+"   throw new Error(\"Invalid index.\");");

      out.println();
      out.println(space.spc+"if (n != null) n.setParent(this);");

      space.updateSpc(-1);
      out.println(space.spc + "}");

      //
      // Output get children method
      //
      out.println();
      out.println(space.spc+"public java.util.Vector getChildren() {");
      
      names = nameList.elements();
      space.updateSpc(+1);

      out.println(space.spc+"java.util.Vector vector = new java.util.Vector("+nameList.size()+");");

      for ( int count = 0; names.hasMoreElements(); ++count ) {
         String name = (String)names.nextElement();
         out.println(space.spc + "vector.addElement("+name+");");
      }

      out.println(space.spc+"return vector;");

      space.updateSpc(-1);
      out.println(space.spc + "}");
      
      //
      // Finish it all off
      //
      space.updateSpc(-1);
      out.println(space.spc + "}");

      out.flush();
      out.close();
      return buf.toString();
   }
}

//
// An extension of VPrinter that prints the comments for the generated Visitor.
// Also shows which field stors which part of the production.
//
// Modified to use JavaDoc style comments "/* */".  Assumes the opening and
// closing comment tokens are printed before and after this class's output.
//
class CommentPrinter extends Printer {
   private boolean visitedLookahead = false;
   private Spacing spc;
   private int nestLevel = 0;
   private int fieldNum = 0;

   CommentPrinter(PrintWriter w, Spacing s) {
      super(w);
      spc = s;
   }

   private String getFieldName() {
      return "f" + String.valueOf(fieldNum++);
   }

   public void visit(NodeToken n) {
      if ( !Globals.javaDocComments ) out.print(n.tokenImage);
      else {
         // Convert special HTML characters
         for ( int i = 0; i < n.tokenImage.length(); ++i ) {
            char c = n.tokenImage.charAt(i);
            if ( c == '<' ) out.print("&lt;");
            else if ( c == '>' ) out.print("&gt;");
            else out.print(c);
         }
      }
   }

   public void visit(LocalLookahead n) {
      visitedLookahead = true;
   }

   public void visit(ExpansionChoices n) {
      if ( !n.f1.present() ) {
         n.f0.accept(this);
         return;
      }

      if ( nestLevel == 0 )
         out.print(spc.spc + " * " + getFieldName() + " -> ");

      ++nestLevel;
      n.f0.accept(this);
      --nestLevel;

      for ( Enumeration e = n.f1.elements(); e.hasMoreElements(); ) {
         NodeSequence seq = (NodeSequence)e.nextElement();

         if ( nestLevel == 0 ) {
            out.println();
            out.print(spc.spc + " *       ");
         }
         else out.print(" ");

         out.print(seq.elementAt(0) + " ");
         ++nestLevel;
         seq.elementAt(1).accept(this);
         --nestLevel;
      }
   }

   public void visit(Expansion n) {
      if ( !n.f0.present() ) return;

      Enumeration e = n.f0.elements();
      ExpansionUnit unit = (ExpansionUnit)e.nextElement();
      visitedLookahead = false;

      if ( nestLevel == 0 && unit.f0.which != 0 && unit.f0.which != 1 )
         out.print(spc.spc + " * " + getFieldName() + " -> ");

      ++nestLevel;
      if ( unit.f0.which != 1) unit.accept(this); // fix for Java block bug
      --nestLevel;

      for ( ; e.hasMoreElements(); ) {
         unit = (ExpansionUnit)e.nextElement();
         if ( !visitedLookahead ) {
            if ( unit.f0.which != 0 && unit.f0.which != 1 ) {
               if ( nestLevel == 0 ) {
                  out.println();
                  out.print(spc.spc + " * " + getFieldName() + " -> ");
               }
               else
                  out.print(" ");
            }
         }
         else visitedLookahead = false;

         ++nestLevel;
         if ( unit.f0.which != 1) unit.accept(this); // fix for Java block bug
         --nestLevel;
      }
   }
}
