/*
 * All files in the distribution of JTB, The Java Tree Builder are
 * Copyright 1997, 1998 by the Purdue Research Foundation of Purdue
 * University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and
 * other materials related to such distribution and use acknowledge
 * that the software was developed at Purdue University, West Lafayette,
 * Indiana by Kevin Tao and Jens Palsberg.  No charge may be made
 * for copies, derivations, or distributions of this material
 * without the express written consent of the copyright holder.
 * Neither the name of the University nor the name of the author
 * may be used to endorse or promote products derived from this
 * material without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 */

//
// Class ClassGenerator creates a vector of ClassInfo objects describing
// every class to be generated.
//
// Programming notes:
//    - do not continue down the tree once a new field has been added to
//      curClass--we only worry about top-level expansions 
//

package EDU.purdue.jtb.visitor;

import EDU.purdue.jtb.syntaxtree.*;
import EDU.purdue.jtb.misc.*;
import java.util.*;
import java.io.*;

public class ClassGenerator extends DepthFirstVisitor {
   private int fieldNum = 0;
   private boolean printToken = false;

   // Variables to store the list of classes generated
   private ClassInfo curClass;
   private Vector classList = new Vector();

   public Vector getClassList()  { return classList; }
   private void resetFieldNum()  { fieldNum = 0; }

   private String curFieldName() {
      return "f" + String.valueOf(fieldNum++);
   }

   //
   // User generated visitor methods below
   //
   //
   // f0 -> JavaCCOptions()
   // f1 -> < PARSER_BEGIN_TK > 
   // f2 -> < LPAREN > 
   // f3 -> < IDENTIFIER > 
   // f4 -> < RPAREN > 
   // f5 -> CompilationUnit()
   // f6 -> < PARSER_END_TK > 
   // f7 -> < LPAREN > 
   // f8 -> < IDENTIFIER > 
   // f9 -> < RPAREN > 
   // f10 -> ( Production() )*
   // f11 -> < EOF > 
   //
   public void visit(JavaCCInput n) {
      n.f10.accept(this);
   }

   //
   // f0 -> < JAVACODE_TK > 
   // f1 -> ResultType()
   // f2 -> < IDENTIFIER > 
   // f3 -> FormalParameters()
   // f4 -> Block()
   //
   public void visit(JavaCodeProduction n) {
      // Don't visit
   }

   //
   // f0 -> ResultType()
   // f1 -> < IDENTIFIER > 
   // f2 -> FormalParameters()
   // f3 -> < COLON > 
   // f4 -> Block()
   // f5 -> < LBRACE > 
   // f6 -> ExpansionChoices()
   // f7 -> < RBRACE > 
   //
   public void visit(BNFProduction n) {
      resetFieldNum();
      printToken = true;
      curClass = new ClassInfo(n.f6, n.f1.toString());
      classList.addElement(curClass);

      n.f6.accept(this);
      printToken = false;
   }

   //
   // f0 -> [ LexicalStateList() ]
   // f1 -> RegExprKind()
   // f2 -> [ < LBRACKET >  < IGNORE_CASE_TK >  < RBRACKET >  ]
   // f3 -> < COLON > 
   // f4 -> < LBRACE > 
   // f5 -> RegExprSpec()
   // f6 -> ( < BIT_OR >  RegExprSpec() )*
   // f7 -> < RBRACE > 
   //
   public void visit(RegularExprProduction n) {
      // Don't visit--don't want to generate NodeTokens inside RegularExpression
      // if it's visited from a RegularExpressionProduction
   }

   //
   // f0 -> < TOKEN_MGR_DECLS_TK > 
   // f1 -> < COLON > 
   // f2 -> Block()
   //
   public void visit(TokenManagerDecls n) {
      // Don't visit
   }

   //
   // f0 -> Expansion()
   // f1 -> ( < BIT_OR >  Expansion() )*
   //
   public void visit(ExpansionChoices n) {
      if ( !n.f1.present() )
         n.f0.accept(this);
      else
         curClass.addField("NodeChoice", curFieldName());
   }

   //
   // f0 -> ( ExpansionUnit() )*
   //
   public void visit(Expansion n) {
      n.f0.accept(this);
   }

   //
   // f0 -> LocalLookahead()
   //       | Block() 
   //       | < LPAREN >  ExpansionChoices() < RPAREN >  [ < PLUS > | < STAR >  | < HOOK >   ] 
   //       | < LBRACKET >  ExpansionChoices() < RBRACKET >  
   //       | [ PrimaryExpression() < ASSIGN >  ] ExpansionUnitTerm() 
   //
   public void visit(ExpansionUnit n) {
      switch ( n.f0.which ) {
         case 0: return;      // Lookahead
         case 1: return;      // Java block
         case 2:              // Parenthesized expansion
            NodeSequence seq = (NodeSequence)n.f0.choice;
            NodeOptional ebnfMod = (NodeOptional)seq.elementAt(3);

            if ( ebnfMod.present() ) {
               NodeChoice modChoice = (NodeChoice)ebnfMod.node;
               String mod = ((NodeToken)modChoice.choice).tokenImage;
               curClass.addField(getNameForMod(mod), curFieldName());
            }
            else {
               ExpansionChoices ec = (ExpansionChoices)seq.elementAt(1);

               if ( ec.f1.present() )
                  curClass.addField("NodeChoice", curFieldName());
               else
                  curClass.addField("NodeSequence", curFieldName());
            }

            break;
         case 3:              // Optional expansion
            curClass.addField("NodeOptional", curFieldName());
            break;
         case 4:              // Normal production
            n.f0.accept(this);
            break;
         default:
            Errors.hardErr("n.f0.which = " + String.valueOf(n.f0.which));
            break;
      }
   }

   private String getNameForMod(String mod) {
      if ( mod.equals("+") )        return "NodeList";
      else if ( mod.equals("*") )   return "NodeListOptional";
      else if ( mod.equals("?") )   return "NodeOptional";
      else {
         Errors.hardErr("Illegal EBNF modifier in " +
                        "ExpansionUnit: mod = " + mod);
         return "";
      }
   }

   //
   // f0 -> RegularExpression()
   //       | < IDENTIFIER >  Arguments() 
   //
   public void visit(ExpansionUnitTerm n) {
      switch ( n.f0.which ) {
         case 0 :
            n.f0.accept(this);
            break;
         case 1 :
            NodeSequence seq = (NodeSequence)n.f0.choice;
            String ident = ((NodeToken)seq.elementAt(0)).tokenImage;
            curClass.addField(ident, curFieldName());
            break;
         default:
            Errors.hardErr("n.f0.which = " + String.valueOf(n.f0.which));
      }
   }

   //
   // f0 -> < LOOKAHEAD_TK > 
   // f1 -> < LPAREN > 
   // f2 -> [ < INTEGER_LITERAL >  ]
   // f3 -> [ < COMMA >  ]
   // f4 -> ExpansionChoices()
   // f5 -> [ < COMMA >  ]
   // f6 -> [ < LBRACE >  Expression() < RBRACE >  ]
   // f7 -> < RPAREN > 
   //
   public void visit(LocalLookahead n) {
      // Don't visit...ignore lookaheads
   }

   //
   // f0 -> < STRING_LITERAL > 
   //       | < LT >  [ [ < POUND >  ] < IDENTIFIER >  < COLON >  ] ComplexRegularExpressionChoices() < GT >  
   //       | < LT >  < IDENTIFIER >  < GT >  
   //       | < LT >  < EOF_TK >  < GT >  
   //
   public void visit(RegularExpression n) {
      if ( printToken )
         curClass.addField("NodeToken", curFieldName());
   }
}
