/*
 * All files in the distribution of JTB, The Java Tree Builder are
 * Copyright 1997, 1998 by the Purdue Research Foundation of Purdue
 * University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and
 * other materials related to such distribution and use acknowledge
 * that the software was developed at Purdue University, West Lafayette,
 * Indiana by Kevin Tao and Jens Palsberg.  No charge may be made
 * for copies, derivations, or distributions of this material
 * without the express written consent of the copyright holder.
 * Neither the name of the University nor the name of the author
 * may be used to endorse or promote products derived from this
 * material without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 */

package EDU.purdue.jtb.visitor;

import EDU.purdue.jtb.syntaxtree.*;
import EDU.purdue.jtb.misc.Errors;
import java.util.*;

/**
 * Semantic checking phase for JTB checks for the following conditions:
 *   - When productions have a return value other than void since JTB
 *     automatically alters the return value of all productions in the
 *     annotated grammar.
 *   - For blocks of Java code within productions since to our knowledge
 *     they are generally unnecessary in JTB grammars.
 *   - When any productions have a name reserved by an automatically
 *     generated JTB class (e.g. Node, NodeList, etc.)
 *   - Extraneous parentheses in a production
 *   - JavaCode productions must be handled specially, as stated in the
 *     JTB Release Notes page.
 */
public class SemanticChecker extends DepthFirstVisitor {
   String prod;            // name of the current production

   /**
    * f0 -> JavaCCOptions()
    * f1 -> <PARSER_BEGIN_TK>
    * f2 -> <LPAREN>
    * f3 -> <IDENTIFIER>
    * f4 -> <RPAREN>
    * f5 -> CompilationUnit()
    * f6 -> <PARSER_END_TK>
    * f7 -> <LPAREN>
    * f8 -> <IDENTIFIER>
    * f9 -> <RPAREN>
    * f10 -> ( Production() )*
    * f11 -> <EOF>
    */
   public void visit(JavaCCInput n) {
      n.f10.accept(this);           // only visit production subtree
   }

   /**
    * f0 -> [ <OPTIONS_TK> <LBRACE> ( OptionBinding() )* <RBRACE> ]
    */
   public void visit(JavaCCOptions n) {
   }

   /**
    * f0 -> ( <IDENTIFIER> | <LOOKAHEAD_TK> | <IGNORE_CASE_TK> | <STATIC> )
    * f1 -> <ASSIGN>
    * f2 -> ( <INTEGER_LITERAL> | BooleanLiteral() | <STRING_LITERAL> )
    * f3 -> <SEMICOLON>
    */
   public void visit(OptionBinding n) {
   }

   /**
    * f0 -> JavaCodeProduction()
    *       | RegularExprProduction()
    *       | BNFProduction()
    *       | TokenManagerDecls()
    */
   public void visit(Production n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <JAVACODE_TK>
    * f1 -> ResultType()
    * f2 -> <IDENTIFIER>
    * f3 -> FormalParameters()
    * f4 -> Block()
    */
   public void visit(JavaCodeProduction n) {
      Errors.warning("Javacode block must be specially handled.  See JTB " +
         "Release Notes web page.", n.f0.beginLine);
   }

   /**
    * f0 -> ResultType()
    * f1 -> <IDENTIFIER>
    * f2 -> FormalParameters()
    * f3 -> <COLON>
    * f4 -> Block()
    * f5 -> <LBRACE>
    * f6 -> ExpansionChoices()
    * f7 -> <RBRACE>
    */
   public void visit(BNFProduction n) {
      prod = n.f1.tokenImage;

      if ( n.f0.f0.which != 0 )           // return type isn't void
         Errors.warning("Non-void return type in " + prod + "().",
            n.f1.beginLine);

      if ( prod.equals("Node") || prod.equals("NodeListInterface") ||
           prod.equals("NodeList") || prod.equals("NodeListOptional") ||
           prod.equals("NodeOptional") || prod.equals("NodeSequence") ||
           prod.equals("NodeToken") || prod.equals("NodeChoice") )
         Errors.softErr("Production \"" + prod + "\" has the same name as a " +
            "JTB-generated class.", n.f1.beginLine);

      n.f6.accept(this);
   }

   /**
    * f0 -> [ LexicalStateList() ]
    * f1 -> RegExprKind()
    * f2 -> [ <LBRACKET> <IGNORE_CASE_TK> <RBRACKET> ]
    * f3 -> <COLON>
    * f4 -> <LBRACE>
    * f5 -> RegExprSpec()
    * f6 -> ( <BIT_OR> RegExprSpec() )*
    * f7 -> <RBRACE>
    */
   public void visit(RegularExprProduction n) {
   }

   /**
    * f0 -> <TOKEN_MGR_DECLS_TK>
    * f1 -> <COLON>
    * f2 -> ClassBodyDeclaration()
    */
   public void visit(TokenManagerDecls n) {
   }

   /**
    * f0 -> <LT> <STAR> <GT>
    *       | <LT> <IDENTIFIER> ( <COMMA> <IDENTIFIER> )* <GT>
    */
   public void visit(LexicalStateList n) {
   }

   /**
    * f0 -> <TOKEN_TK>
    *       | <SPECIAL_TOKEN_TK>
    *       | <SKIP_TK>
    *       | <MORE_TK>
    */
   public void visit(RegExprKind n) {
   }

   /**
    * f0 -> RegularExpression()
    * f1 -> [ Block() ]
    * f2 -> [ <COLON> <IDENTIFIER> ]
    */
   public void visit(RegExprSpec n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> Expansion()
    * f1 -> ( <BIT_OR> Expansion() )*
    */
   public void visit(ExpansionChoices n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * f0 -> ( ExpansionUnit() )*
    */
   public void visit(Expansion n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> LocalLookahead()
    *       | Block()
    *       | <LPAREN> ExpansionChoices() <RPAREN> [ <PLUS> | <STAR> | <HOOK> ]
    *       | <LBRACKET> ExpansionChoices() <RBRACKET>
    *       | [ PrimaryExpression() <ASSIGN> ] ExpansionUnitTerm()
    */
   public void visit(ExpansionUnit n) {
      if ( n.f0.which == 1 )           // unit is a block
         Errors.warning("Block of Java code in " + prod + "().",
            ((Block)n.f0.choice).f0.beginLine);

      if ( n.f0.which == 2 ) {         // parentheses present
         NodeSequence seq = (NodeSequence)n.f0.choice;
         ExpansionChoices choice = (ExpansionChoices)seq.elementAt(1);
         NodeOptional mod = (NodeOptional)seq.elementAt(3);

         if ( !mod.present() && !choice.f1.present() )
            Errors.warning("Extra parentheses in " + prod + "().",
               ((NodeToken)((NodeSequence)n.f0.choice).elementAt(0)).beginLine);
      }

      n.f0.accept(this);
   }

   /**
    * f0 -> RegularExpression()
    *       | <IDENTIFIER> Arguments()
    */
   public void visit(ExpansionUnitTerm n) {
   }

   /**
    * f0 -> <LOOKAHEAD_TK>
    * f1 -> <LPAREN>
    * f2 -> [ <INTEGER_LITERAL> ]
    * f3 -> [ <COMMA> ]
    * f4 -> ExpansionChoices()
    * f5 -> [ <COMMA> ]
    * f6 -> [ <LBRACE> Expression() <RBRACE> ]
    * f7 -> <RPAREN>
    */
   public void visit(LocalLookahead n) {
   }

   /**
    * f0 -> <STRING_LITERAL>
    *       | <LT> [ [ <POUND> ] <IDENTIFIER> <COLON> ] ComplexRegularExpressionChoices() <GT>
    *       | <LT> <IDENTIFIER> <GT>
    *       | <LT> <EOF_TK> <GT>
    */
   public void visit(RegularExpression n) {
   }

   /**
    * f0 -> ComplexRegularExpression()
    * f1 -> ( <BIT_OR> ComplexRegularExpression() )*
    */
   public void visit(ComplexRegularExpressionChoices n) {
   }

   /**
    * f0 -> ( ComplexRegularExpressionUnit() )*
    */
   public void visit(ComplexRegularExpression n) {
   }

   /**
    * f0 -> <STRING_LITERAL>
    *       | <LT> <IDENTIFIER> <GT>
    *       | CharacterList()
    *       | <LPAREN> ComplexRegularExpressionChoices() <RPAREN> [ <PLUS> | <STAR> | <HOOK> ]
    */
   public void visit(ComplexRegularExpressionUnit n) {
   }

   /**
    * f0 -> [ <TILDE> ]
    * f1 -> <LBRACKET>
    * f2 -> [ CharacterDescriptor() ( <COMMA> CharacterDescriptor() )* ]
    * f3 -> <RBRACKET>
    */
   public void visit(CharacterList n) {
   }

   /**
    * f0 -> <STRING_LITERAL>
    * f1 -> [ <MINUS> <STRING_LITERAL> ]
    */
   public void visit(CharacterDescriptor n) {
   }

}
